{
  "cells": [
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "%matplotlib inline"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "# Residual networks\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "There are three different approaches to using BackPACK with ResNets.\n\n1. `Custom ResNet`: (Only works for first-order extensions) Write your own model\n   by defining its forward pass. Trainable parameters must be in modules known to\n   BackPACK (e.g. :class:`torch.nn.Conv2d`, :class:`torch.nn.Linear`).\n\n2. `Custom ResNet with BackPACK custom modules`: (Works for first- and second-\n   order extensions) Build your ResNet with custom modules provided by BackPACK\n   without overwriting the forward pass. This approach is useful if you want to\n   understand how BackPACK handles ResNets, or if you think building a container\n   module that implicitly defines the forward pass is more elegant than coding up\n   a forward pass.\n\n3. `Any ResNet with BackPACK's converter`: (Works for first- and second-order\n   extensions) Convert your model into a BackPACK-compatible architecture.\n\n<div class=\"alert alert-info\"><h4>Note</h4><p>ResNets are still an experimental feature. Always double-check your\n   results, as done in this example! Open an issue if you encounter a bug to help\n   us improve the support.\n\n   Not all extensions support ResNets (yet). Please create a feature request in the\n   repository if the extension you need is not supported.</p></div>\n\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Let's get the imports out of the way.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "from torch import (\n    allclose,\n    cat,\n    cuda,\n    device,\n    int32,\n    linspace,\n    manual_seed,\n    rand,\n    rand_like,\n)\nfrom torch.nn import (\n    Conv2d,\n    CrossEntropyLoss,\n    Flatten,\n    Identity,\n    Linear,\n    Module,\n    MSELoss,\n    ReLU,\n    Sequential,\n)\nfrom torch.nn.functional import cross_entropy, relu\nfrom torchvision.models import resnet18\n\nfrom backpack import backpack, extend\nfrom backpack.custom_module.branching import Parallel, SumModule\nfrom backpack.custom_module.graph_utils import BackpackTracer\nfrom backpack.extensions import BatchGrad, DiagGGNExact\nfrom backpack.utils.examples import autograd_diag_ggn_exact, load_one_batch_mnist\n\nmanual_seed(0)\nDEVICE = device(\"cuda:0\" if cuda.is_available() else \"cpu\")\nx, y = load_one_batch_mnist(batch_size=32)\nx, y = x.to(DEVICE), y.to(DEVICE)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Custom ResNet\nWe can build a ResNet by extending :py:class:`torch.nn.Module`.\nAs long as the layers with parameters (:py:class:`torch.nn.Conv2d`\nand :py:class:`torch.nn.Linear`) are ``nn`` modules, BackPACK can extend them,\nand this is all that is needed for first-order extensions.\nWe can rewrite the :code:`forward` to implement the residual connection,\nand :py:func:`extend() <backpack.extend>` the resulting model.\n\n<div class=\"alert alert-info\"><h4>Note</h4><p>Using in-place operations is not compatible with PyTorch's\n   :meth:`torch.nn.Module.register_full_backward_hook`. Therefore,\n   always use :code:`x = x + residual` instead of :code:`x += residual`.</p></div>\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "class MyFirstResNet(Module):\n    def __init__(self, C_in=1, C_hid=5, input_dim=(28, 28), output_dim=10):\n        \"\"\"Instantiate submodules that are used in the forward pass.\"\"\"\n        super().__init__()\n\n        self.conv1 = Conv2d(C_in, C_hid, kernel_size=3, stride=1, padding=1)\n        self.conv2 = Conv2d(C_hid, C_hid, kernel_size=3, stride=1, padding=1)\n        self.linear1 = Linear(input_dim[0] * input_dim[1] * C_hid, output_dim)\n        if C_in == C_hid:\n            self.shortcut = Identity()\n        else:\n            self.shortcut = Conv2d(C_in, C_hid, kernel_size=1, stride=1)\n\n    def forward(self, x):\n        \"\"\"Manual implementation of the forward pass.\"\"\"\n        residual = self.shortcut(x)\n        x = self.conv2(relu(self.conv1(x)))\n        x = x + residual  # don't use: x += residual\n        x = x.flatten(start_dim=1)\n        x = self.linear1(x)\n        return x\n\n\nmodel = extend(MyFirstResNet()).to(DEVICE)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "The loss does not need to be extended in this case either, as it does not\nhave model parameters and BackPACK does not need to know about it for\nfirst-order extensions. This also means you can use any custom loss function.\n\nUsing :py:class:`BatchGrad <backpack.extensions.BatchGrad>` in a\n:py:class:`with backpack(...) <backpack.backpack>` block,\nwe can access the individual gradients for each sample.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "loss = cross_entropy(model(x), y, reduction=\"sum\")\n\nwith backpack(BatchGrad()):\n    loss.backward()\n\nfor name, parameter in model.named_parameters():\n    print(f\"{name:>20}'s grad_batch shape: {parameter.grad_batch.shape}\")"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "To check that everything works, let's compute one individual gradient with\nPyTorch (using a single sample in a forward and backward pass)\nand compare it with the one computed by BackPACK.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "sample_to_check = 1\nx_to_check = x[[sample_to_check]]\ny_to_check = y[[sample_to_check]]\n\nmodel.zero_grad()\nloss = cross_entropy(model(x_to_check), y_to_check)\nloss.backward()\n\nprint(\"Do the individual gradients match?\")\nfor name, parameter in model.named_parameters():\n    match = allclose(parameter.grad_batch[sample_to_check], parameter.grad, atol=1e-6)\n    print(f\"{name:>20}: {match}\")\n    if not match:\n        raise AssertionError(\"Individual gradients don't match!\")"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Custom ResNet with BackPACK custom modules\nSecond-order extensions only work if every node in the computation graph is an\n``nn`` module that can be extended by BackPACK. The above ResNet class\n:py:class:`MyFirstResNet<MyFirstResNet>` does not satisfy these conditions, because\nit implements the skip connection via :py:func:`torch.add` while overwriting the\n:py:meth:`forward() <torch.nn.Module.forward>` method.\n\nTo build ResNets without overwriting the forward pass, BackPACK offers custom modules:\n\n1. :py:class:`Parallel<backpack.branching.Parallel>` is similar to\n   :py:class:`torch.nn.Sequential`, but implements a container for a parallel sequence\n   of modules (followed by an aggregation module), rather than a sequential one.\n\n2. :py:class:`SumModule<backpack.branching.SumModule>` is the module that takes the\n   role of :py:func:`torch.add` in the previous example. It sums up multiple inputs.\n   We will use it to merge the skip connection.\n\nWith the above modules, we can build a simple ResNet as a container that implicitly\ndefines the forward pass:\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "C_in = 1\nC_hid = 2\ninput_dim = (28, 28)\noutput_dim = 10\n\nmodel = Sequential(\n    Conv2d(C_in, C_hid, kernel_size=3, stride=1, padding=1),\n    ReLU(),\n    Parallel(  # skip connection with ReLU-activated convolution\n        Identity(),\n        Sequential(\n            Conv2d(C_hid, C_hid, kernel_size=3, stride=1, padding=1),\n            ReLU(),\n        ),\n        merge_module=SumModule(),\n    ),\n    Flatten(),\n    Linear(input_dim[0] * input_dim[1] * C_hid, output_dim),\n)\n\nmodel = extend(model.to(DEVICE))\nloss_function = extend(CrossEntropyLoss(reduction=\"mean\")).to(DEVICE)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "This ResNets supports BackPACK's second-order extensions:\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "loss = loss_function(model(x), y)\n\nwith backpack(DiagGGNExact()):\n    loss.backward()\n\nfor name, parameter in model.named_parameters():\n    print(f\"{name}'s diag_ggn_exact: {parameter.diag_ggn_exact.shape}\")\n\ndiag_ggn_exact_vector = cat([p.diag_ggn_exact.flatten() for p in model.parameters()])"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Comparison with :py:mod:`torch.autograd`:\n\n<div class=\"alert alert-info\"><h4>Note</h4><p>Computing the full GGN diagonal with PyTorch's built-in automatic differentiation\n   can be slow, depending on the number of parameters. To reduce run time, we only\n   compare some elements of the diagonal.</p></div>\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "num_params = sum(p.numel() for p in model.parameters())\nnum_to_compare = 10\nidx_to_compare = linspace(0, num_params - 1, num_to_compare, device=DEVICE, dtype=int32)\n\ndiag_ggn_exact_to_compare = autograd_diag_ggn_exact(\n    x, y, model, loss_function, idx=idx_to_compare\n)\n\nprint(\"Do the exact GGN diagonals match?\")\nfor idx, element in zip(idx_to_compare, diag_ggn_exact_to_compare):\n    match = allclose(element, diag_ggn_exact_vector[idx], atol=1e-6)\n    print(f\"Diagonal entry {idx:>6}: {match}\")\n    if not match:\n        raise AssertionError(\"Exact GGN diagonals don't match!\")"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Any ResNet with BackPACK's converter\nIf you are not building a ResNet through custom modules but for instance want to\nuse a prominent ResNet from :py:mod:`torchvision.models`, BackPACK offers a converter.\nIt analyzes the model and tries to turn it into a compatible architecture. The result\nis a :py:class:`torch.fx.GraphModule` that exclusively consists of modules.\n\nHere, we demo the converter on :py:class:`resnet18 <torchvision.models.resnet18>`.\n\n<div class=\"alert alert-info\"><h4>Note</h4><p>:py:class:`resnet18 <torchvision.models.resnet18>` has to be in evaluation mode,\n   because it contains batch normalization layers that are not supported in train\n   mode by the second-order extension used in this example.</p></div>\n\nLet's create the model, and convert it in the call to :py:func:`extend <backpack.extend>`:\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "loss_function = extend(MSELoss().to(DEVICE))\nmodel = resnet18(num_classes=5).to(DEVICE).eval()\n\n# use BackPACK's converter to extend the model (turned off by default)\nmodel = extend(model, use_converter=True)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "To get an understanding what happened, we can inspect the model's graph with the\nfollowing helper function:\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "def print_table(module: Module) -> None:\n    \"\"\"Prints a table of the module.\n\n    Args:\n        module: module to analyze\n    \"\"\"\n    graph = BackpackTracer().trace(module)\n    graph.print_tabular()\n\n\nprint_table(model)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Admittedly, the converted :py:class:`resnet18 <torchvision.models.resnet18>`'s graph\nis quite large. Note however that it fully consists of modules (indicated by\n``call_module`` in the first table column) such that BackPACK's hooks can\nsuccessfully backpropagate additional information for its second-order extensions\n(first-order extensions work, too).\n\nLet's verify that second-order extensions are working:\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "x = rand(4, 3, 7, 7, device=DEVICE)  # (128, 3, 224, 224)\noutput = model(x)\ny = rand_like(output)\n\nloss = loss_function(output, y)\n\nwith backpack(DiagGGNExact()):\n    loss.backward()\n\nfor name, parameter in model.named_parameters():\n    print(f\"{name}'s diag_ggn_exact: {parameter.diag_ggn_exact.shape}\")\n\ndiag_ggn_exact_vector = cat([p.diag_ggn_exact.flatten() for p in model.parameters()])"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Comparison with :py:mod:`torch.autograd`:\n\n<div class=\"alert alert-info\"><h4>Note</h4><p>Computing the full GGN diagonal with PyTorch's built-in automatic differentiation\n   can be slow, depending on the number of parameters. To reduce run time, we only\n   compare some elements of the diagonal.</p></div>\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "num_params = sum(p.numel() for p in model.parameters())\nnum_to_compare = 10\nidx_to_compare = linspace(0, num_params - 1, num_to_compare, device=DEVICE, dtype=int32)\n\ndiag_ggn_exact_to_compare = autograd_diag_ggn_exact(\n    x, y, model, loss_function, idx=idx_to_compare\n)\n\nprint(\"Do the exact GGN diagonals match?\")\nfor idx, element in zip(idx_to_compare, diag_ggn_exact_to_compare):\n    match = allclose(element, diag_ggn_exact_vector[idx], atol=1e-6)\n    print(f\"Diagonal entry {idx:>8}: {match}\")\n    if not match:\n        raise AssertionError(\"Exact GGN diagonals don't match!\")"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.7.9"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}